/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * END COPYRIGHT BLOCK **/
/*
 * ntdaemon.c: Standalone daemon support
 * 
 * Rob McCool
 */


/* The definitions that activate code in this file are defined in <base/systems.h
 * The two code bases have been merged into 1. NSPR can be turned off using a 
 * define in the NT Makefile.
 */

#define DAEMON_STATS 			  
#include <base/systems.h>
/* Removed for ns security integration
#include "sslio/sslio.h"
*/

#ifdef MCC_HTTPD
extern void clf_rotate();
#endif /* MCC_HTTPD */

#ifdef NET_SOCKETS

#define MAX_LIFE 1024
/* The number of processes by which to grow the mob at one time */
#define MOB_GROWSIZE 4

#include "ereport.h"
#include "util.h"
#include <file.h>
#include <nt/magnus.h>


#include <stdio.h>
#include <process.h>    /* beginthread, endthread */
#include "base/eventlog.h"
#include "base/daemon.h"
#include "frame/conf.h"
#include "safs/ntlogbuf.h" /* MLM for logbuf_cleanup_all */
#include <nt/messages.h>
#include <nt/regparms.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#ifdef  NET_SSL
#include "sec.h"
#include "ssl.h"
int secure;
#endif	/* NET_SSL */

static void (*chcb)(Session *);
static void (*rcb)(int);
static void (*rotatecb)();
static int mobport, mobmaxthreads, mobminthreads, mobthreads, moblife;

#define DAEMON_STATS

#define MOB_RESTART 0
#define MOB_SUICIDE 1
#define MOB_GROW 2
#define MOB_RESPAWN 3

#define EVENTS 4
#define MOVELOG 4

static HANDLE Events[5];

#ifdef MCC_ADMSERV // disable body of daemon in async/accelerator build

/* These are the functions that are used to respond
 * to mob events */

typedef void MobFunction();
typedef MobFunction *MobFunctionPointer;

MobFunction _mob_restart;
MobFunction _mob_suicide;
MobFunction _mob_grow;
MobFunction _mob_respawn;
MobFunction _movelog;

static MobFunctionPointer MobFunctionTable[] = {
	_mob_restart,
	_mob_suicide,
	_mob_grow,
	_mob_respawn,
	_movelog
};

#endif  // MCC_ADMSERV

 
typedef struct {
	int port;
	int slot;
	SYS_NETFD sd;
	char address[64];
	int AddressNumber;
} child_argument;

static BOOLEAN Terminator;
static int AddressNumber;
static SYS_NETFD growfd;
//char *server_id;

#if defined (USE_NSPR)
#include <base/crit.h>
#include <base/systhr.h>
static CRITICAL Exit;
static CONDVAR MainExitVariable;
#endif /* !USE_NSPR  */

/* Proxy needs this to be global */
int child_mode;
multiple_domains *array;
static DWORD dwTlsIndex;

BOOLEAN StartHttpdService(LPCTSTR serviceName);

/* defined in nterrors.c . This is the only place where this is called. */
extern "C" void HashNtErrors();

#define PRINT_EXIT_WARNING()\
    ((void)MessageBox(GetDesktopWindow(), \
    "The server is terminating due to an error." \
    "Check the event viewer for the error message", \
    "SERVER EXITING!", MB_ICONEXCLAMATION | MB_OK))

#define EXIT_RETURN() \
	_mob_suicide(); \
	ereport_terminate(); \
    PRINT_EXIT_WARNING(); \
	exit(1);

#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
#define TERMINATE_THREAD systhread_terminate(systhread_current());
#else /* !USE_NSPR || NO_NSPR_THREADS */
#define TERMINATE_THREAD _endthreadex(0);
#endif /* !USE_NSPR || NO_NSPR_THREADS */



/* Use an exit variable to synchronize cleanup of the service */

NSAPI_PUBLIC VOID
InitializeExitVariable()
{
#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
	Exit = crit_init();
	MainExitVariable = condvar_init(Exit);
#endif /* !USE_NSPR || NO_NSPR_THREADS */
}

NSAPI_PUBLIC VOID
WaitForExitVariable(void)
{
#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
	/* Set the MainThread's Monitor so we will know when it
	 * goes away . Used by service termination procedures */
	condvar_wait(MainExitVariable);
	crit_terminate(Exit);
#endif /* !USE_NSPR || NO_NSPR_THREADS */
}

/* kill all the service threads. Used by service termination procedures
 * to do cleanup */

NSAPI_PUBLIC BOOL
kill_child_threads()
{
#ifdef MCC_ADMSERV
	logbuf_cleanup_all();
    if (SetEvent(Events[MOB_SUICIDE]) == FALSE) {
        return FALSE;
    } else {
        return TRUE;
    }
#else /* !MCC_ADMSERV */
	return TRUE;
#endif /* !MCC_ADMSERV */
}

/*
 * kill the server. This procedure can be used to terminate the server at any
 * time from within the server.
 */

VOID
kill_server()
{
    HANDLE server_handle;

    ereport(LOG_INFORM,"Terminating the server %s", server_id);

    server_handle = OpenEvent(EVENT_ALL_ACCESS, FALSE, server_id);

    if (server_handle == NULL) {
        ereport(LOG_FAILURE,"kill_server:cannot open server event %s", 
            system_errmsg());
        return;
    }

    if(!SetEvent(server_handle)) {
        ereport(LOG_FAILURE,"kill_server:cannot set server event %s", 
            system_errmsg());
        return;
    }
}

#ifdef MCC_BATMAN
/*  Stub these out for batman.  */

void child_status(int mode)
{
}

void child_status2(int http_status, long bytes)
{
}
#endif

/* stats is set to indicate what the processes are doing */
int *stats;

#ifdef MCC_ADMSERV // disable body of daemon in async/accelerator build
void child_status(int mode)
{
#ifdef MCC_ADMSERV
	int statslot;
	LPVOID lpvData;

	lpvData = TlsGetValue(dwTlsIndex);
	statslot = *(DWORD *)lpvData;

	if (stats)
		stats[statslot] = mode;
#endif
}

void child_status2(int http_status, long bytes)
{
    /* 01-30-95 MB- so far the child_status2 is not implemented on NT? */
    return;
}



/* 
 * Creates a listener socket on the given port. Used by UNIX and NT daemons.
 */

SYS_NETFD _create_listener(char *ipstr, int port)
{
    SYS_NETFD sd;
    int one=1;
    struct sockaddr_in sa_server;
	int optionValue = SO_SYNCHRONOUS_ALERT;

    if ((sd = net_socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) == SYS_NET_ERRORFD) {
        ereport(LOG_FAILURE,"error: could not get socket (%s)\n",
                system_errmsg());
		EXIT_RETURN();
    }

    if((net_setsockopt(sd, SOL_SOCKET, SO_REUSEADDR,
                       (void *)&one, sizeof(one))) == -1) 
    {
        ereport(LOG_FAILURE,"error: could not set socket option (%s)\n",
                system_errmsg());
		EXIT_RETURN();
    }
    ZERO((char *) &sa_server, sizeof(sa_server));
    sa_server.sin_family=AF_INET;
    sa_server.sin_addr.s_addr=(ipstr ? inet_addr(ipstr) : htonl(INADDR_ANY));
    sa_server.sin_port = htons((u_short)port);
    if(net_bind(sd, (struct sockaddr *) &sa_server,sizeof(sa_server)) == -1) {
		if (ipstr) {
	        ereport(LOG_FAILURE,"Terminating Service:error: "
	        	"could not bind to address %s port %d (%s)\n", 
	        		ipstr, port, system_errmsg());
			EXIT_RETURN();
		} else {
	        ereport(LOG_FAILURE,"Terminating Service:error: could not bind to port %d (%s)\n", port,
	                system_errmsg());
	 		EXIT_RETURN();
		}
    }
    net_listen(sd, 100);

    return sd;
}


/* These parameters are set by the main thread and passed to child threads */
typedef struct {
    SYS_NETFD csd;
	struct sockaddr_in sa_client;
	int cl;
} _connection_s;

/* These should only be written once. */
/* MLM - redefinition error when building admserv */
#ifndef MCC_ADMSERV
static void (*chcb)(Session *);
#endif
static HANDLE mpsem;

/* Counters for parent process respawn */
static unsigned long num_exits, num_forks;

/* statfd is an open file pointer to the shared data file */
static HANDLE statfd;

#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
static SYS_THREAD *nspr_threads;
#else /* !USE_NSPR || NO_NSPR_THREADS */
/* statpids keeps a list of which pid has which stat spot */
static  DWORD *statpids;
static HANDLE *stathandles;
#endif /* !USE_NSPR || NO_NSPR_THREADS */


SYS_NETFD
SetHandleNonInheritable(SYS_NETFD csd)
{
   BOOL ret;
   DWORD error;
   HANDLE hDupSocket;

   ret = SetHandleInformation((HANDLE)PR_FileDesc2NativeHandle(csd), HANDLE_FLAG_INHERIT, 0);

   if(ret)
        return csd;
    
    /* This is probably a Win95 system */

    error = GetLastError();
    if (error == ERROR_CALL_NOT_IMPLEMENTED) {

        /* We now need to use the Duplicate socket trick to make the
         * socket non-inheritable. First Duplicate csd as newSocket.
         * Then call SSL_ReplaceSocket to close the old socket, 
         * and attach socket information to the new socket...*/

    	if (! DuplicateHandle(GetCurrentProcess(), (HANDLE)PR_FileDesc2NativeHandle(csd), 
    	        GetCurrentProcess(), &hDupSocket, 0, FALSE, 
    	        DUPLICATE_SAME_ACCESS)) {
        		ereport(LOG_FAILURE, "SetHandleNonInheritable:" 
        		    "could not duplicate socket (%s)", system_errmsg());
                return csd;
    	}

    	PR_ChangeFileDescNativeHandle(csd, SSL_ReplaceSocket(PR_FileDesc2NativeHandle(csd), (int)hDupSocket));
	return csd;
    }

    /* We have a failure condition here that is not because the
     * SetHandleInformation call is not implemented. */

	ereport(LOG_FAILURE, 
      "Could not SetHandleInformation (%s)", system_errmsg());

    return csd;
}


void _mob_initstats(int port, int maxthreads)
{
    char fn[256], *growme, tmpdir[512];
    int statsize;
    PCHAR  statmap;
    int count;
	HANDLE fileMapHandle, fileHandle;
	DWORD BytesWritten;
	
	/* XXX I'm not sure that DWORD still needs to be added. Depends upon
	 * sitemon. What about char vs int for UNIX/NT */

    statsize = sizeof(DWORD) + (16 * maxthreads * sizeof(int));

	strcpy(tmpdir, "\\temp");
	CreateDirectory("\\temp", NULL);

    sprintf(fn, "%s\\daemonstat.%s", tmpdir, server_id);
	if((fileHandle = CreateFile(fn, GENERIC_READ | GENERIC_WRITE, 
			FILE_SHARE_READ,
			NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 
			NULL)) == INVALID_HANDLE_VALUE) {
				goto disable;
	}
    growme = (char *) MALLOC(statsize);
    ZERO(growme, statsize);
    if(WriteFile (fileHandle, (char *)growme, statsize, 
    		&BytesWritten, NULL) == FALSE) {
        goto disable;
    }
    FREE(growme);

	if ((fileMapHandle = CreateFileMapping(fileHandle, NULL, PAGE_READWRITE,
			0, 0, NULL)) == NULL) {
		CloseHandle(fileHandle);
		goto disable;
	}
	if (!(statmap = (char *)MapViewOfFile(fileMapHandle, FILE_MAP_WRITE, 0, 0, 0))) {
		CloseHandle(fileMapHandle);
		CloseHandle(fileHandle);
		goto disable;
	}

    stats = (int *)((char *) statmap + sizeof(DWORD));
#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
	/* What we have is an array of pointers to SYS_THREAD structures */
	nspr_threads = (SYS_THREAD *)MALLOC(16 * maxthreads * (sizeof(SYS_THREAD)));
    for(count = 0; count < maxthreads * 16; count++) {
        stats[count] = CHILD_EMPTY_SLOT;
		nspr_threads[count] = NULL;
    }
#else /* !USE_NSPR || NO_NSPR_THREADS */
    statpids = (DWORD *) MALLOC(16 * maxthreads * sizeof(DWORD));
    stathandles = (HANDLE *) MALLOC(16 * maxthreads * sizeof(HANDLE));
    for(count = 0; count < maxthreads * 16; count++) {
        stats[count] = CHILD_EMPTY_SLOT;
        statpids[count] = (DWORD)-1;
		stathandles[count] = 0;
    }
#endif /* !USE_NSPR || NO_NSPR_THREADS */

    return;

  disable:
    ereport(LOG_FAILURE, "Terminating Service:Failure: Could not open statistics file (%s)\n", 
    	system_errmsg());
    stats = NULL;
	EXIT_RETURN();
}


int dns_init(void);

#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
void _child_loop(void *argument)
#else /* !USE_NSPR || NO_NSPR_THREADS */
DWORD WINAPI _child_loop(LPVOID argument)
#endif /* !USE_NSPR || NO_NSPR_THREADS */
{
   SYS_NETFD csd;
   struct sockaddr_in sa_client;
   Session *sn;
   int x, error, clen, life, child_mode;
   child_argument *arg = (child_argument *)argument;
   int port = arg->port;
   int statslot = arg->slot;
   int Address = arg->AddressNumber;
   SYS_NETFD sd = arg->sd;
   int threads = array[Address].threads;
   LPVOID lpvData;
   DWORD tid = GetCurrentThreadId();

#ifdef MALLOC_POOLS
   /*
    * For each session we will build a dynamic memory pool and then
    * tear it down when the session is closed.  The problem is that
    * each thread needs to have its own pool, and there is no good
    * place to store a pointer that is safe except in thread private
    * data.   For now, initialize to NULL, so memory will be alloc'd
    * outside of any pool.
    */
   systhread_setdata(getThreadMallocKey(), NULL);
#endif

#ifdef NET_SSL
   SEC_ResetRandom();
#endif

#if defined (USE_NSPR) && !defined(NET_WINSOCK)
    dns_init();  /* XXX error check? */
#endif /* USE_NSPR */

   life = moblife;
   lpvData = (LPVOID)MALLOC(sizeof(DWORD));
   *(DWORD *)(lpvData) = statslot;

   if (!TlsSetValue(dwTlsIndex, lpvData)) {
        ereport(LOG_WARN, "Could not set Thread Local Storage Value"
        		" for thread at slot %d", statslot); 
   }

   while(1) {
       child_mode = 0;
       clen=sizeof(sa_client);
	   child_status(CHILD_AWAIT_CONNECT);

       while(1) {
		   if(Terminator) {
                TERMINATE_THREAD;
		return;
		   }

		   threads = array[Address].threads;
           csd = net_accept(sd, (struct sockaddr *) &sa_client, &clen);
           if(csd != SYS_NET_ERRORFD)
                break;

#ifdef NET_SSL
           if(secure)
               ereport(LOG_WARN, "secure handshake failed (code %d)\n",
                       XP_GetError());
           else
#endif /* NET_SSL */
				{
					error = WSAGetLastError();
					if (error == WSANOTINITIALISED ||
						error == WSAEINTR)	{
						net_close(sd);
						FREE(argument);
						FREE(lpvData);
						goto FINISH;
					} else {
               		 	ereport(LOG_WARN, "accept failed %d (%s)", 
               				error, system_winsockerr());
					}
				}
        }  /* while(1) accept loop */

	    /* SetHandleNonInheritable takes the socket handle and makes it
	     * non-inheritable */

	    csd = SetHandleNonInheritable(csd);

	    child_mode = 1;
	    child_status(CHILD_PROCESSING);
	    if (stats) {
	       for(x = Address * mobmaxthreads; x < (Address * mobmaxthreads + threads) ; ++x)
	           if(stats[x] == CHILD_AWAIT_CONNECT)
	               break;
        }
        if(x == (Address * mobmaxthreads + threads))	{
	   		growfd = sd;
			AddressNumber = Address;
			if(SetEvent(Events[MOB_GROW]) == FALSE) {
				ereport(LOG_WARN, "Failed to pulse Event %d %s",
				MOB_GROW, system_errmsg());
			} else {
				//ereport(LOG_INFORM,"growing address %d", AddressNumber);
			}
	   }
	   
       sn = session_create(csd, &sa_client);
       if (sn) {
           ZeroMemory(sn->inbuf->address, 64);
           if (arg->address) {
               strcpy(sn->inbuf->address ,arg->address);
           }
           (chcb)(sn);
           session_free(sn);
       }
	   		
       if((child_mode == 2) || (!--life)) {
	        if(stats) {
	            for(x = Address * mobmaxthreads; x < (Address * mobmaxthreads + threads) ; x++) {
#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
					if (nspr_threads[x] == 	systhread_current()) {
						InterlockedExchange((long *)&stats[x], CHILD_EMPTY_SLOT);
						nspr_threads[x] = NULL;
						break;
					}
#else /* !USE_NSPR || NO_NSPR_THREADS */
	                if(statpids[x] == tid) {
	                    InterlockedExchange(&statpids[x], -1);
	                    InterlockedExchange((LPLONG)&stathandles[x], (LONG)0);
	                    InterlockedExchange(&stats[x], CHILD_EMPTY_SLOT);
                        break;
	                }
#endif /* !USE_NSPR || NO_NSPR_THREADS */
	            }
	        }
	        if((!stats) || (mobthreads < mobminthreads))
	            ++child_mode;
			AddressNumber = Address;
			if(SetEvent(Events[MOB_RESPAWN]) == FALSE) {
				ereport(LOG_FAILURE, "Failed to send MobGrowth Event to parent %s",
					system_errmsg());
			} else {
    			InterlockedDecrement((long *)&mobthreads);
				InterlockedDecrement((long *)&(array[AddressNumber].threads));
				ereport(LOG_INFORM,"Pulsing MobRespawn Event %d", AddressNumber);
			}
			FREE(lpvData);
			FREE(argument);
            TERMINATE_THREAD;
	    return;
		}  /* if child_mode == 2 || (!--life) */
   }  /* while (1) */
FINISH:
   return;
}

void _mob_respawn()
{

	// Hmmm. Maybe cost of waking up server to do this >
	// cost of locking and incrementing and decrementing...

	if((!stats) || (mobthreads < mobminthreads))
            ++num_exits;
        ereport(LOG_INFORM, "respawn thread pool to %d (%d)", 
                mobthreads, AddressNumber);
}

void _movelog()
{
	HANDLE EventHandle;

	(*rotatecb)();
	if (!(EventHandle = OpenEvent(EVENT_ALL_ACCESS, FALSE, "NetscapeMoveLog"))) {
		ereport(LOG_FAILURE,"Could not open event to signal rotate application"
			"Could not create the MoveLog event:%s", system_errmsg());
		return;
	}

	if(SetEvent(EventHandle) == FALSE) {
		ereport(LOG_FAILURE, "Failed to send MoveLog Event to rotate app %s",
			system_errmsg());
	}
	 
}

void _mob_restart()
{
	// Kill the children and then call the restart function
	Terminator = TRUE;
    (*rcb)(0);
}

void _mob_suicide()
{
    char fn[1024], tmpdir[512];
	int i = 0;
	BOOLEAN socketsClosed = FALSE, threadTerminated = FALSE;
	DWORD result;

	Terminator = TRUE;
	// Kill the children
	for (i = 0; i < (16 * mobmaxthreads); i++) {
#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
		if(nspr_threads[i]) { 
			systhread_terminate(nspr_threads[i]);
			return;
		}
#else /* !USE_NSPR || NO_NSPR_THREADS */
		if (stathandles[i]) {
			threadTerminated = TerminateThread(stathandles[i], 0);
			CloseHandle(stathandles[i]);
		}
#endif /* !USE_NSPR || NO_NSPR_THREADS */
	}
	result = GetEnvironmentVariable("temp",
		tmpdir, 512);
	if (result > 512 || result == 0) {
		strcpy(tmpdir, "\\temp");
	}	
    sprintf(fn, "%s\\daemonstat.%s", tmpdir, server_id);
    if(stats) {
        CloseHandle (statfd);
        DeleteFile(fn);
#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
		FREE(nspr_threads);
#else /* !USE_NSPR || NO_NSPR_THREADS */
        FREE(statpids);
		FREE(stathandles);
#endif /* !USE_NSPR || NO_NSPR_THREADS */

    }
	WSACleanup();
#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
	/* Notify its parent that it is done */
    if(MainExitVariable)
    	condvar_notify(MainExitVariable);
#endif /* !USE_NSPR || NO_NSPR_THREADS */
    TERMINATE_THREAD;
}

void _mob_grow()
{
    register unsigned int x;
    static time_t lastgrowth = 0;
    time_t tt;
	
    tt = time(NULL);
    if(mobthreads < mobmaxthreads) {
        x = (unsigned int) (mobmaxthreads - mobthreads);
        if(x > MOB_GROWSIZE)
            x = MOB_GROWSIZE;
        num_exits += x;
        ereport(LOG_INFORM, "growing thread pool from %d to %d", 
                mobthreads, mobthreads + x);
    }
}

BOOLEAN
StartHttpdService(LPCTSTR serviceName)
{
	SC_HANDLE schSCManager, schService;
	DWORD Error;

	if (!(schSCManager = OpenSCManager(
                        NULL,                   // machine (NULL == local)
                        NULL,                   // database (NULL == default)
                        SC_MANAGER_ALL_ACCESS   // access required
                        ))) {
		ereport(LOG_INFORM,
			 "Could not open the ServiceControlManager, Error %d", 
			 GetLastError());
		return FALSE;
	}
    schService = OpenService(
        				schSCManager,               // SCManager database
        				serviceName,                // name of service
						SERVICE_ALL_ACCESS);

    if (schService == NULL) {
        ereport(LOG_INFORM, 
        	"StartNetsiteService:Could not open the service %s: Error %d",
        	serviceName, GetLastError());
		CloseServiceHandle(schService);
        return FALSE;
    } 
	if (!StartService(schService, 0, NULL)) {
		Error = GetLastError();
		if (Error == ERROR_SERVICE_ALREADY_RUNNING) {
			CloseHandle(schService);
			CloseHandle(schSCManager);
			return TRUE;
		}
		ereport(LOG_INFORM, 
			"StartNetsiteService:Could not start the service %s",
			serviceName, Error);
		CloseServiceHandle(schService);
		return FALSE;
	}
	return TRUE;
}


void ntdaemon_init(daemon_s *d, SECURITY_ATTRIBUTES *saSecurity)
{
	PSECURITY_DESCRIPTOR pSD = NULL;
	WSADATA wsd;
    int i, y = 0;
	multiple_domains domain;
	CHAR ErrMsg[256];

    server_id = STRDUP(d->servid);

	memset ((VOID *)saSecurity, 0, sizeof(SECURITY_ATTRIBUTES));
	if (!(pSD = (PSECURITY_DESCRIPTOR)MALLOC(
			SECURITY_DESCRIPTOR_MIN_LENGTH))) {
		ereport(LOG_WARN,"Service Startup: Could not allocate "
			" security descriptor");
        goto CONTINUE;

	}
	if (!InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION)) {
		ereport(LOG_WARN,"Service Startup: Could not init"
			" security descriptor");
        goto CONTINUE;
	}

	if(!SetSecurityDescriptorDacl(pSD, TRUE, (PACL)NULL, FALSE)) {
		ereport(LOG_WARN,"Service Startup: Could not set the"
			" security Dacl");
        goto CONTINUE;
	}

	saSecurity->nLength = sizeof(SECURITY_ATTRIBUTES);
	saSecurity->bInheritHandle = TRUE;
	saSecurity->lpSecurityDescriptor = pSD;

CONTINUE:
	if(WSAStartup(MAKEWORD(1, 1), &wsd) != 0) {
	    sprintf( ErrMsg, "WSAstartup failed (%s)\n", system_winsockerr());
	    LogErrorEvent(NULL, EVENTLOG_ERROR_TYPE, 0,
	    	MSG_BAD_WINSOCK_INIT, ErrMsg, NULL );
		ereport(LOG_INFORM,"Terminating Service:WinSock init failed: %s",ErrMsg);
        PRINT_EXIT_WARNING();
		exit(1);
	}
	// d->maxthreads should be at least equal to the number of domains...
	// How many domains do we have ?

	y = 0;
	for (i = 0; i < 16; i++) {
		domain = d -> domains[i];
		if ( domain.address) {
			y++;
		}
	}
	if (d->maxthreads < y)
		d->maxthreads = y;
			
    _mob_initstats(d->port, d->maxthreads);
	mobport = d->port;
	mobmaxthreads = d->maxthreads;
	mobminthreads = (d->minthreads == -1 ? mobmaxthreads : d->minthreads);
	moblife = (d->proclife == -1 ? MAX_LIFE : d->proclife);

#ifdef NET_SSL
    secure = d->security;
#endif

    if(stats)
        num_exits = mobminthreads;
    else
    num_exits = d->maxthreads;

    mobthreads = 0;
	Terminator = FALSE;
	num_forks = 0;
	array = d->domains;

	if ((dwTlsIndex = TlsAlloc()) == 0xFFFFFFFF) {
	    sprintf( ErrMsg, "TlsAlloc failed (%s)\n", system_errmsg());
	    LogErrorEvent(NULL, EVENTLOG_ERROR_TYPE, 0,
	    	MSG_BAD_STARTUP, ErrMsg, NULL );
		ereport(LOG_INFORM,"Httpd Server Startup failed: %s",ErrMsg);
	}

    chcb = d->child_callback;
	rcb = d->rcback;
	rotatecb = d->rotate_callback;



    /* set up error handling */
    system_errmsg_init();
}

void ntdaemon_start_initthread(daemon_s *d,	multiple_domains domain, 
    SYS_NETFD fd, int Address)
{
    int statslot;
    child_argument *args;
    unsigned ThreadId;
    unsigned long ThreadHandle;
    
    MainExitVariable = 0;

#ifdef MALLOC_POOLS
    setThreadMallocKey(systhread_newkey());
#endif

    if(stats) {
        for(statslot = Address * mobmaxthreads; 
        	statslot < (Address * mobmaxthreads) + mobmaxthreads; statslot++) {
#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
			if(nspr_threads[statslot] == NULL)
#else /* !USE_NSPR || NO_NSPR_THREADS */
          if(statpids[statslot] == -1)
#endif /* !USE_NSPR || NO_NSPR_THREADS */
				break;
		}
        if(statslot == (Address * mobmaxthreads + mobmaxthreads ) ) {
            ereport(LOG_WARN, "can't find empty statistics slot");
            statslot = Address * mobmaxthreads;
        }
    }
	args = (child_argument *)MALLOC(sizeof (child_argument));
	args->port = d->port;
	args->slot = statslot;
	args->sd = fd;
	args->AddressNumber = Address;
    if (domain.address) {
    	strcpy(args->address, domain.address);
    }

 #if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
    nspr_threads[statslot] = systhread_start(15, 65536, _child_loop, 
                                         (void *)(args));
    if(stats) {
            InterlockedExchange((long *)&stats[statslot], CHILD_AWAIT_CONNECT);
    }
#else /* !USE_NSPR || NO_NSPR_THREADS */
	if ((ThreadHandle = (VOID *)_beginthreadex( NULL, 0, _child_loop, 
		(void *)(args), 0, &ThreadId)) == 0) {
			ereport(LOG_INFORM, "NT daemon: could not create new thread %d",
				errno);
	}
    if(stats) {
            InterlockedExchange(&statpids[statslot] , ThreadId);
            InterlockedExchange((LPLONG)&stathandles[statslot] , (LONG)ThreadHandle);
            InterlockedExchange(&stats[statslot], CHILD_AWAIT_CONNECT);
    }
#endif /* !USE_NSPR || NO_NSPR_THREADS */
    ++mobthreads;
    ++num_forks;
}

#if defined (USE_NSPR) && !defined (NO_NSPR_THREADS)
void _ntdaemon_run(void *argument)
#else /* !USE_NSPR || NO_NSPR_THREADS */
unsigned int __stdcall _ntdaemon_run(void *argument)
#endif /* !USE_NSPR || NO_NSPR_THREADS */
{
	int i, y = 0, Address;
	CHAR ErrMsg[256];
	DWORD WaitResult;
	multiple_domains domain;
	SECURITY_ATTRIBUTES saSecurity;
	daemon_s *d = (daemon_s *)argument;

    net_init(d->security);

    ntdaemon_init(d, &saSecurity);

	/* First check to see if no addresses have been specified.
	 * If this is true, then jump to safety_net.
	 */

	if (!(d -> domains[0].address)) {
		goto SAFETY_NET;
	}


	/* First assign at least one thread per address over-riding 
	 * minthreads if necessary */

	for (i = 0; i < 16; i++) {
		domain = d -> domains[i];
		if ( domain.address) {
			d->domains[i].sd = domain.sd = _create_listener(domain.address, d->port);
            if(d->security) {
                SSL_ENABLE(domain.sd, SSL_SECURITY, 1);
		SSL_ENABLE(domain.sd, SSL_ENABLE_SSL3, d->ssl3);
		SSL_ENABLE(domain.sd, SSL_ENABLE_SSL2, d->ssl2);
		SSL_ENABLE(domain.sd, SSL_DELAYED_HANDSHAKE, 1);
                if(d->secure_auth)
                    SSL_ENABLE(domain.sd, SSL_REQUEST_CERTIFICATE, 1);
            }

			Address = domain.AddressNumber;
            ntdaemon_start_initthread(d, domain, domain.sd, Address);
            d->domains[i].threads++;
		}
	}
	/* If minthreads > num_forks then distribute the remaining threads among the
	 * addresses */

	if (num_exits < num_forks) {
		num_exits = num_forks;
	} else while (num_exits > num_forks) {

		for (i = 0; i < 16; i++) {
			domain = d -> domains[i];
			Address = domain.AddressNumber;
			if ( domain.sd) {
                ntdaemon_start_initthread(d, domain, domain.sd, Address);
				d->domains[i].threads++;
				if ( num_forks == num_exits)
					break;
			} /* if domain[i].sd */
		} /* for (i = 0; i < 16; i++) */
	}  /* num_exits < num_forks */

SAFETY_NET:
	if ( num_exits != num_forks) {
		growfd = _create_listener(NULL, d->port);
        if(d->security) {
            SSL_ENABLE(growfd, SSL_SECURITY, 1);
	    SSL_ENABLE(growfd, SSL_ENABLE_SSL3, d->ssl3);
	    SSL_ENABLE(growfd, SSL_ENABLE_SSL2, d->ssl2);
            if(d->secure_auth)
                SSL_ENABLE(growfd, SSL_REQUEST_CERTIFICATE, 1);
        }

		d->domains[0].sd = growfd;
		AddressNumber = d->domains[0].AddressNumber;
	} else {
		/* For safety's sake */
		growfd = d->domains[0].sd;
		AddressNumber = 1;
	}
	 
	for (i = 0; i < EVENTS; i++) {
		if(!(Events[i] = CreateEvent(NULL, FALSE, FALSE, NULL))) {
			ereport(LOG_CATASTROPHE,"Service Startup Failure"
				"Terminating Service:Could not create event %d:%s", i ,system_errmsg()); 
			EXIT_RETURN();
		}
	}

	sprintf(ErrMsg,"%s.MoveLog", d->servid);
	if (rotatecb) {
		// Now create the Last Named Event
		if (!(Events[MOVELOG] = CreateEvent(&saSecurity, FALSE, FALSE, ErrMsg))) {
			ereport(LOG_FAILURE,"Service Startup Error"
				"Could not create the MoveLog event:%s", system_errmsg());
		}
	}

	while(1) {
		while(num_exits != num_forks) {
            /* Don't fork if at process limit. */
            if(mobthreads >= mobmaxthreads) {
                ++num_forks;
                continue;
            }
           ntdaemon_start_initthread(d, array[AddressNumber], growfd, AddressNumber);
           array[AddressNumber].threads++;
		} /* while child_mode */
		if (rotatecb) {
			if ((WaitResult = WaitForMultipleObjects( 5, Events, FALSE, INFINITE)) 
				== WAIT_FAILED) {
				ereport(LOG_CATASTROPHE, "Failed to wait on Event objects %s",
					system_errmsg());
			} else {
				(*MobFunctionTable[WaitResult - WAIT_OBJECT_0])();
				ResetEvent(Events[WaitResult - WAIT_OBJECT_0]);
			}
		} else {
			if ((WaitResult = WaitForMultipleObjects( 4, Events, FALSE, INFINITE)) 
				== WAIT_FAILED) {
				ereport(LOG_CATASTROPHE, "Failed to wait on Event objects %s",
					system_errmsg());
			} else {
				(*MobFunctionTable[WaitResult - WAIT_OBJECT_0])();
				ResetEvent(Events[WaitResult - WAIT_OBJECT_0]);
			}
		}
	}	/* while (1) */
	return;
	
}
#endif  /* NET_SOCKETS */

#endif // MCC_ADMSERV

